/*T
   Concepts: KSP^solving a system of linear equations
   Concepts: KSP^Laplacian, 2d
   Processors: n
T*/

/*
Added at the request of Marc Garbey.

Inhomogeneous Laplacian in 2D. Modeled by the partial differential equation

   div \rho grad u = f,  0 < x,y < 1,

   with forcing function

   f = e^{-x^2/\nu} e^{-y^2/\nu}

   with Dirichlet boundary conditions

   u = f(x,y) for x = 0, x = 1, y = 0, y = 1

   or pure Neumman boundary conditions

   This uses multigrid to solve the linear system
 */

static char help[] = "Solves 2D inhomogeneous Laplacian using multigrid.\n\n";

#include <petscdmda.h>
#include <petscksp.h>
//#include <petscpcmg.h>
#include <petsc.h>
#include "ComputeFunctions.h"
#include "GlobalVariables.h"

extern PetscErrorCode ComputeMatrix(KSP,Mat,Mat,MatStructure*, void*);
extern PetscErrorCode ComputePpRHS(KSP, Vec, void*);
extern PetscErrorCode Initialization(DM, int);
extern PetscErrorCode BuildNonUniformGrid();
extern PetscErrorCode ApplyNonUniformGrid(DM);
extern PetscErrorCode ComputeInitialGuess(KSP, Vec, void*);

#undef __FUNCT__
#define __FUNCT__ "main"
int main(int argc,char **argv) {
  FILE *fpCp;
  int time = 0, TotalTime;
  DM             da;
  PetscReal      norm;
  PetscErrorCode ierr;
  PetscInt       stages[7];
  KSP	ksp;
  Vec	x,b,r;
  Mat	A;

  PetscInitialize(&argc,&argv,(char *)0,help);
  MPI_Comm_rank(PETSC_COMM_WORLD, &rank);
  MPI_Comm_size(PETSC_COMM_WORLD, &NumRank);
  if(!rank) { printf("NumRank = %d \n", NumRank);} else{}
  sprintf(fname, "%s%d%s", "Results/SurfaceCp", rank, ".dat");
  fpCp = fopen(fname, "w");

  ierr = PetscLogStageRegister("DMMG Setup",&stages[0]);  CHKERRQ(ierr);
  ierr = PetscLogStageRegister("Pressure RHS Setup",&stages[1]);  CHKERRQ(ierr);
  ierr = PetscLogStageRegister("Pressure Solve",&stages[2]);  CHKERRQ(ierr);
  ierr = PetscLogStageRegister("Corrector RHS Setup",&stages[3]);  CHKERRQ(ierr);
  ierr = PetscLogStageRegister("Corrector Solve",&stages[4]);  CHKERRQ(ierr);

  ierr = PetscLogStagePush(stages[0]);CHKERRQ(ierr);   // Start DMMG Setup
  PetscPrintf(PETSC_COMM_WORLD, "Ready for BuildingNonUniformGrid!!\n");
  ierr = BuildNonUniformGrid();
  ierr = KSPCreate(PETSC_COMM_WORLD,&ksp);CHKERRQ(ierr);
  ierr = DMDACreate3d(PETSC_COMM_WORLD,DMDA_BOUNDARY_NONE,DMDA_BOUNDARY_NONE,DMDA_BOUNDARY_NONE,DMDA_STENCIL_STAR,-imax,-jmax,-kmax,PETSC_DECIDE,PETSC_DECIDE,PETSC_DECIDE,1,1,0,0,0,&da);CHKERRQ(ierr);
  ierr = DMDASetUniformCoordinates(da, 0., (imax-1)*dx, 0., (jmax-1)*dy, 0, (kmax-1)*dz);CHKERRQ(ierr);
  PetscPrintf(PETSC_COMM_WORLD, "Ready for ApplyNonUniformGrid#1!!\n");
  ierr = ApplyNonUniformGrid(da);


  ierr = PetscLogStagePop();  CHKERRQ(ierr); // Finish DMMG Setup
  PetscPrintf(PETSC_COMM_WORLD, "Ready for Initialization!!\n"); 
  ierr = Initialization(da, time);
  ierr = KSPDestroy(&ksp);CHKERRQ(ierr);
  ierr = DMDestroy(&da);CHKERRQ(ierr);
  PetscPrintf(PETSC_COMM_WORLD, "Ready for the main iteration!!\n");


    // -- I -- Solve Poisson Equation

    ierr = PetscLogStagePush(stages[1]);CHKERRQ(ierr);   // Start Pressure RHS Setup
    ierr = KSPCreate(PETSC_COMM_WORLD,&ksp);CHKERRQ(ierr);
    ierr = DMDACreate3d(PETSC_COMM_WORLD,DMDA_BOUNDARY_NONE,DMDA_BOUNDARY_NONE,DMDA_BOUNDARY_NONE,DMDA_STENCIL_STAR,-imax,-jmax,-kmax,PETSC_DECIDE,PETSC_DECIDE,PETSC_DECIDE,1,1,0,0,0,&da);CHKERRQ(ierr);
    ierr = DMDASetUniformCoordinates(da, 0., (imax-1)*dx, 0., (jmax-1)*dy, 0, (kmax-1)*dz);CHKERRQ(ierr);
    PetscPrintf(PETSC_COMM_WORLD, "Ready for ApplyNonUniform of the KSPSolve#1!!\n");
    ierr = ApplyNonUniformGrid(da);

    ierr = KSPSetComputeInitialGuess(ksp,ComputeInitialGuess,PETSC_NULL);CHKERRQ(ierr);
    PetscPrintf(PETSC_COMM_WORLD, "Ready for ComputePpRHS of the KSPSolve#1!!\n");
    ierr = KSPSetComputeRHS(ksp,ComputePpRHS,PETSC_NULL);CHKERRQ(ierr);
    PetscPrintf(PETSC_COMM_WORLD, "Ready for ComputeMatrix of the KSPSolve#1!!\n");
    ierr = KSPSetComputeOperators(ksp,ComputeMatrix,PETSC_NULL);CHKERRQ(ierr); 
    PetscPrintf(PETSC_COMM_WORLD, "Ready for KspSetDM of the KSPSolve#1!!\n");
    ierr = KSPSetDM(ksp,da);CHKERRQ(ierr);
    PetscPrintf(PETSC_COMM_WORLD, "Ready for KSPSetFromOptions of the KSPSolve#1!!\n");
    ierr = KSPSetFromOptions(ksp);CHKERRQ(ierr);
    PetscPrintf(PETSC_COMM_WORLD, "Ready for KSPSetUp of the KSPSolve#1!!\n");
    ierr = KSPSetUp(ksp);CHKERRQ(ierr);
    PetscPrintf(PETSC_COMM_WORLD, "Finish KSPSetUp of the KSPSolve#1!!\n");
 
    ierr = PetscLogStagePop();  CHKERRQ(ierr); // Finish Pressure RHS Setup 
    ierr = PetscLogStagePush(stages[2]);  CHKERRQ(ierr); // Start Pressure Poisson Equation Solve
    PetscPrintf(PETSC_COMM_WORLD, "Ready for KSPSolve#1!!\n");
    ierr = KSPSolve(ksp,PETSC_NULL,PETSC_NULL);CHKERRQ(ierr);
    PetscPrintf(PETSC_COMM_WORLD, "Finish KSPSolve#1!!\n");
    ierr = PetscLogStagePop();  CHKERRQ(ierr); // Finish Pressure Poisson Equation  Solve 
    ierr = KSPGetSolution(ksp,&x);CHKERRQ(ierr);
    ierr = KSPGetRhs(ksp,&b);CHKERRQ(ierr);
    ierr = VecDuplicate(b,&r);CHKERRQ(ierr);
    ierr = KSPGetOperators(ksp,&A,PETSC_NULL,PETSC_NULL);CHKERRQ(ierr);

    ierr = MatMult(A,x,r);CHKERRQ(ierr);
    ierr = VecAXPY(r,-1.0,b);CHKERRQ(ierr);
    ierr = VecNorm(r,NORM_2,&norm);CHKERRQ(ierr);
    ierr = PetscPrintf(PETSC_COMM_WORLD,"Residual norm %G\n",norm);CHKERRQ(ierr);
    ierr = KSPGetDM(ksp, &da);CHKERRQ(ierr);
    ierr = DMDestroy(&da);CHKERRQ(ierr);
    ierr = VecDestroy(&r);CHKERRQ(ierr);
    ierr = KSPDestroy(&ksp);CHKERRQ(ierr);



  ierr = PetscFinalize();



  return 0;
}

#undef __FUNCT__
#define __FUNCT__ "ComputeInitialGuess"
PetscErrorCode ComputeInitialGuess(KSP ksp,Vec b, void *ctx)
{
  PetscErrorCode ierr;

  PetscFunctionBegin;
  ierr = VecSet(b,0);CHKERRQ(ierr);
  PetscFunctionReturn(0);
}


#undef __FUNCT__
#define __FUNCT__ "ComputeMatrix"
PetscErrorCode ComputeMatrix(KSP ksp,Mat jac,Mat B,MatStructure *stflg,void *ctx)
{
  DM             da;
  PetscErrorCode ierr;
  PetscScalar    v[7],Hx,Hy,Hz,HxHydHz,HyHzdHx,HxHzdHy;
  MatStencil     row,col[7];
  PetscInt num, numi, numj, numk;
  double Max1=0., Max2=0., Max3=0.;

  ierr = KSPGetDM(ksp,&da);CHKERRQ(ierr); 
  ierr = DMDAGetInfo(da,0,&mx,&my,&mz,0,0,0,0,0,0,0,0,0);CHKERRQ(ierr);
//  Hx = 1.0 / (PetscReal)(mx-1); Hy = 1.0 / (PetscReal)(my-1); Hz = 1.0 / (PetscReal)(mz-1);
  ierr = DMDAGetCorners(da,&xs,&ys,&zs,&xm,&ym,&zm);CHKERRQ(ierr);

  for (k=zs; k<zs+zm; k++){
    for (j=ys; j<ys+ym; j++){
      for(i=xs; i<xs+xm; i++){
        Hx = PhyDx[i];
        Hy = PhyDy[j];
        Hz = PhyDz[k];

        HxHydHz = Hx*Hy/Hz; HxHzdHy = Hx*Hz/Hy; HyHzdHx = Hy*Hz/Hx;

        row.i = i; row.j = j; row.k = k;
//        if(i == mx-1 || i==0 || j==0 || k==0 || k==mz-1 || j==my-1) {
        if(i == mx-1) {
//        v[0] = 2.0*(HxHydHz + HxHzdHy + HyHzdHx);
          v[0] = 1.;
          ierr = MatSetValuesStencil(B,1,&row,1,&row,v,INSERT_VALUES);CHKERRQ(ierr);
        }
        else if (i==0 || j==0 || k==0 || k==mz-1 || j==my-1) {
          num = 0; numi=0; numj=0; numk=0;
          if (k!=0)
          {
            v[num] = -HxHydHz;
            col[num].i = i;
            col[num].j = j;
            col[num].k = k-1;
            num++; numk++;
          } else {}
          if (j!=0)
          {
            v[num] = -HxHzdHy;
            col[num].i = i;
            col[num].j = j-1;
            col[num].k = k;
            num++; numj++;
          } else {}
          if (i!=0)
          {
            v[num] = -HyHzdHx;
            col[num].i = i-1;
            col[num].j = j;
            col[num].k = k;
            num++; numi++;
          } else {}
          if (i!=mx-1)
          {
            v[num] = -HyHzdHx;
            col[num].i = i+1;
            col[num].j = j;
            col[num].k = k;
            num++; numi++;
          } else {}
          if (j!=my-1)
          {
            v[num] = -HxHzdHy;
            col[num].i = i;
            col[num].j = j+1;
            col[num].k = k;
            num++; numj++;
          } else {}
          if (k!=mz-1)
          {
            v[num] = -HxHydHz;
            col[num].i = i;
            col[num].j = j;
            col[num].k = k+1;
            num++; numk++;
          } else {}
          v[num]   = (PetscReal)(numk)*HxHydHz + (PetscReal)(numj)*HxHzdHy + (PetscReal)(numi)*HyHzdHx;
          col[num].i = i;   col[num].j = j;   col[num].k = k;
          num++;
          ierr = MatSetValuesStencil(jac,1,&row,num,col,v,INSERT_VALUES);CHKERRQ(ierr);
        }

        else {
          v[0] = -HxHydHz;col[0].i = i; col[0].j = j; col[0].k = k-1;
          v[1] = -HxHzdHy;col[1].i = i; col[1].j = j-1; col[1].k = k;
          v[2] = -HyHzdHx;col[2].i = i-1; col[2].j = j; col[2].k = k;
          v[3] = 2.0*(HxHydHz + HxHzdHy + HyHzdHx);col[3].i = row.i; col[3].j = row.j; col[3].k = row.k;
          v[4] = -HyHzdHx;col[4].i = i+1; col[4].j = j; col[4].k = k;
          v[5] = -HxHzdHy;col[5].i = i; col[5].j = j+1; col[5].k = k;
          v[6] = -HxHydHz;col[6].i = i; col[6].j = j; col[6].k = k+1;
          ierr = MatSetValuesStencil(B,1,&row,7,col,v,INSERT_VALUES);CHKERRQ(ierr);
        }
      }
    }
  }
  ierr = MatAssemblyBegin(B,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
  ierr = MatAssemblyEnd(B,MAT_FINAL_ASSEMBLY);CHKERRQ(ierr);
  *stflg = SAME_NONZERO_PATTERN;
  return 0;
}

